@@ -4,7 +4,7 @@
 /*                                                                         */
 /*    Glyph hinter (body).                                                 */
 /*                                                                         */
-/*  Copyright 2000-2001 Catharon Productions Inc.                          */
+/*  Copyright 2000-2001, 2002 Catharon Productions Inc.                    */
 /*  Author: David Turner                                                   */
 /*                                                                         */
 /*  This file is part of the Catharon Typography Project and shall only    */
@@ -25,13 +25,12 @@
 #include "ahangles.h"
 #include "aherrors.h"
 #include FT_OUTLINE_H
-#include FT_INTERNAL_DEBUG_H
 
 
-#define FACE_GLOBALS( face )  ((AH_Face_Globals*)(face)->autohint.data)
+#define FACE_GLOBALS( face )  ((AH_Face_Globals)(face)->autohint.data)
 
 #define AH_USE_IUP
-
+#define OPTIM_STEM_SNAP
 
   /*************************************************************************/
   /*************************************************************************/
@@ -41,7 +40,6 @@
   /*************************************************************************/
   /*************************************************************************/
 
-
   /* snap a given width in scaled coordinates to one of the */
   /* current standard widths                                */
   static FT_Pos
@@ -52,6 +50,7 @@
     int     n;
     FT_Pos  best      = 64 + 32 + 2;
     FT_Pos  reference = width;
+    FT_Pos  scaled;
 
 
     for ( n = 0; n < count; n++ )
@@ -71,16 +70,16 @@
       }
     }
 
+    scaled = (reference+32) & -64;
+
     if ( width >= reference )
     {
-      width -= 0x21;
-      if ( width < reference )
+      if ( width < scaled + 48 )
         width = reference;
     }
     else
     {
-      width += 0x21;
-      if ( width > reference )
+      if ( width > scaled - 48 )
         width = reference;
     }
 
@@ -88,78 +87,143 @@
   }
 
 
-  /* align one stem edge relative to the previous stem edge */
-  static void
-  ah_align_linked_edge( AH_Hinter*  hinter,
-                        AH_Edge*    base_edge,
-                        AH_Edge*    stem_edge,
-                        int         vertical )
+  /* compute the snapped width of a given stem */
+  static FT_Pos
+  ah_compute_stem_width( AH_Hinter  hinter,
+                         int        vertical,
+                         FT_Pos     width )
   {
-    FT_Pos       dist    = stem_edge->opos - base_edge->opos;
-    AH_Globals*  globals = &hinter->globals->scaled;
-    FT_Pos       sign    = 1;
+    AH_Globals  globals = &hinter->globals->scaled;
+    FT_Pos      dist    = width;
+    FT_Int      sign    = 0;
 
 
     if ( dist < 0 )
     {
-      dist = -dist;
-      sign = -1;
+      dist = -width;
+      sign = 1;
     }
 
-    if ( vertical )
+    if ( (  vertical && !hinter->do_vert_snapping ) ||
+         ( !vertical && !hinter->do_horz_snapping ) )
     {
-      dist = ah_snap_width( globals->heights, globals->num_heights, dist );
-
-      /* in the case of vertical hinting, always round */
-      /* the stem heights to integer pixels            */
-      if ( dist >= 64 )
-        dist = ( dist + 16 ) & -64;
-      else
+      /* smooth hinting process, very lightly quantize the stem width */
+      /*                                                              */
+      if ( dist < 64 )
         dist = 64;
+
+      {
+        FT_Pos  delta = dist - globals->stds[vertical];
+
+
+        if ( delta < 0 )
+          delta = -delta;
+
+        if ( delta < 40 )
+        {
+          dist = globals->stds[vertical];
+          if ( dist < 48 )
+            dist = 48;
+        }
+
+        if ( dist < 3 * 64 )
+        {
+          delta = ( dist & 63 );
+          dist &= -64;
+
+          if ( delta < 10 )
+            dist += delta;
+
+          else if ( delta < 32 )
+            dist += 10;
+
+          else if ( delta < 54 )
+            dist += 54;
+
+          else
+            dist += delta;
+        }
+        else
+          dist = ( dist + 32 ) & -64;
+      }
     }
     else
     {
-      dist = ah_snap_width( globals->widths,  globals->num_widths, dist );
-
-      if ( hinter->flags & ah_hinter_monochrome )
+      /* strong hinting process, snap the stem width to integer pixels */
+      /*                                                               */
+      if ( vertical )
       {
-        /* monochrome horizontal hinting: snap widths to integer pixels */
-        /* with a different threshold                                   */
-        if ( dist < 64 )
-          dist = 64;
+        dist = ah_snap_width( globals->heights, globals->num_heights, dist );
+
+        /* in the case of vertical hinting, always round */
+        /* the stem heights to integer pixels            */
+        if ( dist >= 64 )
+          dist = ( dist + 16 ) & -64;
         else
-          dist = ( dist + 32 ) & -64;
+          dist = 64;
       }
       else
       {
-        /* for horizontal anti-aliased hinting, we adopt a more subtle */
-        /* approach: we strengthen small stems, round stems whose size */
-        /* is between 1 and 2 pixels to an integer, otherwise nothing  */
-        if ( dist < 48 )
-          dist = ( dist + 64 ) >> 1;
+        dist = ah_snap_width( globals->widths,  globals->num_widths, dist );
 
-        else if ( dist < 128 )
-          dist = ( dist + 42 ) & -64;
+        if ( hinter->flags & AH_HINTER_MONOCHROME )
+        {
+          /* monochrome horizontal hinting: snap widths to integer pixels */
+          /* with a different threshold                                   */
+          if ( dist < 64 )
+            dist = 64;
+          else
+            dist = ( dist + 32 ) & -64;
+        }
         else
-          /* XXX: round otherwise, prevent color fringes in LCD mode */
-          dist = ( dist + 32 ) & -64;
+        {
+          /* for horizontal anti-aliased hinting, we adopt a more subtle */
+          /* approach: we strengthen small stems, round stems whose size */
+          /* is between 1 and 2 pixels to an integer, otherwise nothing  */
+          if ( dist < 48 )
+            dist = ( dist + 64 ) >> 1;
+
+          else if ( dist < 128 )
+            dist = ( dist + 22 ) & -64;
+          else
+            /* XXX: round otherwise, prevent color fringes in LCD mode */
+            dist = ( dist + 32 ) & -64;
+        }
       }
     }
 
-    stem_edge->pos = base_edge->pos + sign * dist;
+    if ( sign )
+      dist = -dist;
+
+    return dist;
+  }
+
+
+  /* align one stem edge relative to the previous stem edge */
+  static void
+  ah_align_linked_edge( AH_Hinter  hinter,
+                        AH_Edge    base_edge,
+                        AH_Edge    stem_edge,
+                        int        vertical )
+  {
+    FT_Pos  dist = stem_edge->opos - base_edge->opos;
+
+
+    stem_edge->pos = base_edge->pos +
+                     ah_compute_stem_width( hinter, vertical, dist );
   }
 
 
   static void
-  ah_align_serif_edge( AH_Hinter*  hinter,
-                       AH_Edge*    base,
-                       AH_Edge*    serif,
-                       int         vertical )
+  ah_align_serif_edge( AH_Hinter  hinter,
+                       AH_Edge    base,
+                       AH_Edge    serif,
+                       int        vertical )
   {
     FT_Pos  dist;
     FT_Pos  sign = 1;
 
-    UNUSED( hinter );
+    FT_UNUSED( hinter );
 
 
     dist = serif->opos - base->opos;
@@ -169,8 +233,12 @@
       sign = -1;
     }
 
+#if 0
+    if ( dist < 32 )
+      dist = 32;
+#else
     /* do not strengthen serifs */
-    if ( base->flags & ah_edge_done )
+    if ( base->flags & AH_EDGE_DONE )
     {
       if ( dist >= 64 )
         dist = (dist+8) & -64;
@@ -180,6 +248,7 @@
       else
         dist = 0;
     }
+#endif
 
     serif->pos = base->pos + sign * dist;
   }
@@ -198,12 +267,12 @@
 
   /* Another alternative edge hinting algorithm */
   static void
-  ah_hint_edges_3( AH_Hinter*  hinter )
+  ah_hint_edges_3( AH_Hinter  hinter )
   {
-    AH_Edge*     edges;
-    AH_Edge*     edge_limit;
-    AH_Outline*  outline = hinter->glyph;
-    FT_Int       dimension;
+    AH_Edge     edges;
+    AH_Edge     edge_limit;
+    AH_Outline  outline = hinter->glyph;
+    FT_Int      dimension;
 
 
     edges      = outline->horz_edges;
@@ -211,15 +280,15 @@
 
     for ( dimension = 1; dimension >= 0; dimension-- )
     {
-      AH_Edge*  edge;
-      AH_Edge*  anchor = 0;
-      int       has_serifs = 0;
+      AH_Edge  edge;
+      AH_Edge  anchor = 0;
+      int      has_serifs = 0;
 
 
-      if ( ah_debug_disable_vert && !dimension )
+      if ( !hinter->do_horz_hints && !dimension )
         goto Next_Dimension;
 
-      if ( ah_debug_disable_horz && dimension )
+      if ( !hinter->do_vert_hints && dimension )
         goto Next_Dimension;
 
       /* we begin by aligning all stems relative to the blue zone */
@@ -228,11 +297,11 @@
       {
         for ( edge = edges; edge < edge_limit; edge++ )
         {
-          FT_Pos*  blue;
-          AH_Edge  *edge1, *edge2;
+          FT_Pos*      blue;
+          AH_EdgeRec  *edge1, *edge2;
 
 
-          if ( edge->flags & ah_edge_done )
+          if ( edge->flags & AH_EDGE_DONE )
             continue;
 
           blue  = edge->blue_edge;
@@ -254,12 +323,12 @@
             continue;
 
           edge1->pos    = blue[0];
-          edge1->flags |= ah_edge_done;
+          edge1->flags |= AH_EDGE_DONE;
 
           if ( edge2 && !edge2->blue_edge )
           {
             ah_align_linked_edge( hinter, edge1, edge2, dimension );
-            edge2->flags |= ah_edge_done;
+            edge2->flags |= AH_EDGE_DONE;
           }
 
           if ( !anchor )
@@ -271,10 +340,10 @@
       /* relative order of stems in the glyph..                    */
       for ( edge = edges; edge < edge_limit; edge++ )
       {
-        AH_Edge  *edge2;
+        AH_EdgeRec*  edge2;
 
 
-        if ( edge->flags & ah_edge_done )
+        if ( edge->flags & AH_EDGE_DONE )
           continue;
 
         /* skip all non-stem edges */
@@ -287,54 +356,81 @@
 
         /* now, align the stem */
 
-        /* this should not happen, but it's better to be safe.. */
+        /* this should not happen, but it's better to be safe. */
         if ( edge2->blue_edge || edge2 < edge )
         {
 
 #if 0
-          printf( "strange blue alignement, edge %d to %d\n",
+          printf( "strange blue alignment, edge %d to %d\n",
                   edge - edges, edge2 - edges );
 #endif
 
           ah_align_linked_edge( hinter, edge2, edge, dimension );
-          edge->flags |= ah_edge_done;
+          edge->flags |= AH_EDGE_DONE;
           continue;
         }
 
         {
+#if 0
           FT_Bool  min = 0;
-          FT_Pos   delta;
+#endif
+
 
           if ( !anchor )
           {
             edge->pos = ( edge->opos + 32 ) & -64;
             anchor    = edge;
+
+            edge->flags |= AH_EDGE_DONE;
+
+            ah_align_linked_edge( hinter, edge, edge2, dimension );
           }
           else
-            edge->pos = anchor->pos +
-                        ( ( edge->opos - anchor->opos + 32 ) & -64 );
-
-          edge->flags |= ah_edge_done;
-
-          if ( edge > edges && edge->pos < edge[-1].pos )
           {
-            edge->pos = edge[-1].pos;
-            min       = 1;
-          }
+            FT_Pos   org_pos, org_len, org_center, cur_len;
+            FT_Pos   cur_pos1, cur_pos2, delta1, delta2;
 
-          ah_align_linked_edge( hinter, edge, edge2, dimension );
-          delta = 0;
-          if ( edge2 + 1 < edge_limit        &&
-               edge2[1].flags & ah_edge_done )
-            delta = edge2[1].pos - edge2->pos;
 
-          if ( delta < 0 )
-          {
-            edge2->pos += delta;
-            if ( !min )
-              edge->pos += delta;
+            org_pos    = anchor->pos + (edge->opos - anchor->opos);
+            org_len    = edge2->opos - edge->opos;
+            org_center = org_pos + ( org_len >> 1 );
+
+            cur_len    = ah_compute_stem_width( hinter, dimension, org_len );
+
+            cur_pos1   = ( org_pos + 32 ) & -64;
+            delta1     = ( cur_pos1 + ( cur_len >> 1 ) - org_center );
+            if ( delta1 < 0 )
+              delta1 = -delta1;
+
+            cur_pos2   = ( ( org_pos + org_len + 32 ) & -64 ) - cur_len;
+            delta2     = ( cur_pos2 + ( cur_len >> 1 ) - org_center );
+            if ( delta2 < 0 )
+              delta2 = -delta2;
+
+            edge->pos  = ( delta1 <= delta2 ) ? cur_pos1 : cur_pos2;
+            edge2->pos = edge->pos + cur_len;
+
+            edge->flags  |= AH_EDGE_DONE;
+            edge2->flags |= AH_EDGE_DONE;
+
+            if ( edge > edges && edge->pos < edge[-1].pos )
+              edge->pos = edge[-1].pos;
+
+#if 0
+            delta = 0;
+            if ( edge2 + 1 < edge_limit        &&
+                 edge2[1].flags & AH_EDGE_DONE )
+              delta = edge2[1].pos - edge2->pos;
+
+            if ( delta < 0 )
+            {
+              edge2->pos += delta;
+              if ( !min )
+                edge->pos += delta;
+            }
+            edge2->flags |= AH_EDGE_DONE;
+#endif
           }
-          edge2->flags |= ah_edge_done;
         }
       }
 
@@ -345,13 +441,11 @@
       /* to complete our processing                                 */
       for ( edge = edges; edge < edge_limit; edge++ )
       {
-        if ( edge->flags & ah_edge_done )
+        if ( edge->flags & AH_EDGE_DONE )
           continue;
 
         if ( edge->serif )
-        {
           ah_align_serif_edge( hinter, edge->serif, edge, dimension );
-        }
         else if ( !anchor )
         {
           edge->pos = ( edge->opos + 32 ) & -64;
@@ -361,13 +455,13 @@
           edge->pos = anchor->pos +
                       ( ( edge->opos-anchor->opos + 32 ) & -64 );
 
-        edge->flags |= ah_edge_done;
+        edge->flags |= AH_EDGE_DONE;
 
         if ( edge > edges && edge->pos < edge[-1].pos )
           edge->pos = edge[-1].pos;
 
         if ( edge + 1 < edge_limit        &&
-             edge[1].flags & ah_edge_done &&
+             edge[1].flags & AH_EDGE_DONE &&
              edge->pos > edge[1].pos      )
           edge->pos = edge[1].pos;
       }
@@ -379,39 +473,14 @@
   }
 
 
-  FT_LOCAL_DEF void
-  ah_hinter_hint_edges( AH_Hinter*  hinter,
-                        FT_Bool     no_horz_edges,
-                        FT_Bool     no_vert_edges )
+  FT_LOCAL_DEF( void )
+  ah_hinter_hint_edges( AH_Hinter  hinter )
   {
-#if 0
-    ah_debug_disable_horz = no_horz_edges;
-    ah_debug_disable_vert = no_vert_edges;
-#else
-    FT_UNUSED( no_horz_edges );
-    FT_UNUSED( no_vert_edges );    
-#endif
     /* AH_Interpolate_Blue_Edges( hinter ); -- doesn't seem to help      */
     /* reduce the problem of the disappearing eye in the `e' of Times... */
     /* also, creates some artifacts near the blue zones?                 */
     {
       ah_hint_edges_3( hinter );
-
-#if 0
-      /* outline optimizer removed temporarily */
-      if ( hinter->flags & ah_hinter_optimize )
-      {
-        AH_Optimizer  opt;
-
-
-        if ( !AH_Optimizer_Init( &opt, hinter->glyph, hinter->memory ) )
-        {
-          AH_Optimizer_Compute( &opt );
-          AH_Optimizer_Done( &opt );
-        }
-      }
-#endif
-
     }
   }
 
@@ -427,12 +496,12 @@
   /*************************************************************************/
 
   static void
-  ah_hinter_align_edge_points( AH_Hinter*  hinter )
+  ah_hinter_align_edge_points( AH_Hinter  hinter )
   {
-    AH_Outline*  outline = hinter->glyph;
-    AH_Edge*     edges;
-    AH_Edge*     edge_limit;
-    FT_Int       dimension;
+    AH_Outline  outline = hinter->glyph;
+    AH_Edge     edges;
+    AH_Edge     edge_limit;
+    FT_Int      dimension;
 
 
     edges      = outline->horz_edges;
@@ -440,7 +509,7 @@
 
     for ( dimension = 1; dimension >= 0; dimension-- )
     {
-      AH_Edge*   edge;
+      AH_Edge  edge;
 
 
       edge = edges;
@@ -448,12 +517,12 @@
       {
         /* move the points of each segment     */
         /* in each edge to the edge's position */
-        AH_Segment*  seg = edge->first;
+        AH_Segment  seg = edge->first;
 
 
         do
         {
-          AH_Point*  point = seg->first;
+          AH_Point  point = seg->first;
 
 
           for (;;)
@@ -461,12 +530,12 @@
             if ( dimension )
             {
               point->y      = edge->pos;
-              point->flags |= ah_flag_touch_y;
+              point->flags |= AH_FLAG_TOUCH_Y;
             }
             else
             {
               point->x      = edge->pos;
-              point->flags |= ah_flag_touch_x;
+              point->flags |= AH_FLAG_TOUCH_X;
             }
 
             if ( point == seg->last )
@@ -488,15 +557,15 @@
 
   /* hint the strong points -- this is equivalent to the TrueType `IP' */
   static void
-  ah_hinter_align_strong_points( AH_Hinter*  hinter )
+  ah_hinter_align_strong_points( AH_Hinter  hinter )
   {
-    AH_Outline*  outline = hinter->glyph;
-    FT_Int       dimension;
-    AH_Edge*     edges;
-    AH_Edge*     edge_limit;
-    AH_Point*    points;
-    AH_Point*    point_limit;
-    AH_Flags     touch_flag;
+    AH_Outline  outline = hinter->glyph;
+    FT_Int      dimension;
+    AH_Edge     edges;
+    AH_Edge     edge_limit;
+    AH_Point    points;
+    AH_Point    point_limit;
+    AH_Flags    touch_flag;
 
 
     points      = outline->points;
@@ -504,12 +573,12 @@
 
     edges       = outline->horz_edges;
     edge_limit  = edges + outline->num_hedges;
-    touch_flag  = ah_flag_touch_y;
+    touch_flag  = AH_FLAG_TOUCH_Y;
 
     for ( dimension = 1; dimension >= 0; dimension-- )
     {
-      AH_Point*  point;
-      AH_Edge*   edge;
+      AH_Point  point;
+      AH_Edge   edge;
 
 
       if ( edges < edge_limit )
@@ -525,7 +594,8 @@
 #ifndef AH_OPTION_NO_WEAK_INTERPOLATION
           /* if this point is candidate to weak interpolation, we will  */
           /* interpolate it after all strong points have been processed */
-          if ( point->flags & ah_flag_weak_interpolation )
+          if (  ( point->flags & AH_FLAG_WEAK_INTERPOLATION ) &&
+               !( point->flags & AH_FLAG_INFLECTION )         )
             continue;
 #endif
 
@@ -562,8 +632,8 @@
 
           /* otherwise, interpolate the point in between */
           {
-            AH_Edge*  before = 0;
-            AH_Edge*  after  = 0;
+            AH_Edge  before = 0;
+            AH_Edge  after  = 0;
 
 
             for ( edge = edges; edge < edge_limit; edge++ )
@@ -609,7 +679,7 @@
 
       edges      = outline->vert_edges;
       edge_limit = edges + outline->num_vedges;
-      touch_flag = ah_flag_touch_x;
+      touch_flag = AH_FLAG_TOUCH_X;
     }
   }
 
@@ -617,12 +687,12 @@
 #ifndef AH_OPTION_NO_WEAK_INTERPOLATION
 
   static void
-  ah_iup_shift( AH_Point*  p1,
-                AH_Point*  p2,
-                AH_Point*  ref )
+  ah_iup_shift( AH_Point  p1,
+                AH_Point  p2,
+                AH_Point  ref )
   {
-    AH_Point*  p;
-    FT_Pos     delta = ref->u - ref->v;
+    AH_Point  p;
+    FT_Pos    delta = ref->u - ref->v;
 
 
     for ( p = p1; p < ref; p++ )
@@ -634,17 +704,17 @@
 
 
   static void
-  ah_iup_interp( AH_Point*  p1,
-                 AH_Point*  p2,
-                 AH_Point*  ref1,
-                 AH_Point*  ref2 )
+  ah_iup_interp( AH_Point  p1,
+                 AH_Point  p2,
+                 AH_Point  ref1,
+                 AH_Point  ref2 )
   {
-    AH_Point*  p;
-    FT_Pos     u;
-    FT_Pos     v1 = ref1->v;
-    FT_Pos     v2 = ref2->v;
-    FT_Pos     d1 = ref1->u - v1;
-    FT_Pos     d2 = ref2->u - v2;
+    AH_Point  p;
+    FT_Pos    u;
+    FT_Pos    v1 = ref1->v;
+    FT_Pos    v2 = ref2->v;
+    FT_Pos    d1 = ref1->u - v1;
+    FT_Pos    d2 = ref2->u - v2;
 
 
     if ( p1 > p2 )
@@ -703,14 +773,14 @@
 
   /* interpolate weak points -- this is equivalent to the TrueType `IUP' */
   static void
-  ah_hinter_align_weak_points( AH_Hinter*  hinter )
+  ah_hinter_align_weak_points( AH_Hinter  hinter )
   {
-    AH_Outline*  outline = hinter->glyph;
-    FT_Int       dimension;
-    AH_Point*    points;
-    AH_Point*    point_limit;
-    AH_Point**   contour_limit;
-    AH_Flags     touch_flag;
+    AH_Outline  outline = hinter->glyph;
+    FT_Int      dimension;
+    AH_Point    points;
+    AH_Point    point_limit;
+    AH_Point*   contour_limit;
+    AH_Flags    touch_flag;
 
 
     points      = outline->points;
@@ -718,18 +788,18 @@
 
     /* PASS 1: Move segment points to edge positions */
 
-    touch_flag = ah_flag_touch_y;
+    touch_flag = AH_FLAG_TOUCH_Y;
 
     contour_limit = outline->contours + outline->num_contours;
 
-    ah_setup_uv( outline, ah_uv_oy );
+    ah_setup_uv( outline, AH_UV_OY );
 
     for ( dimension = 1; dimension >= 0; dimension-- )
     {
-      AH_Point*   point;
-      AH_Point*   end_point;
-      AH_Point*   first_point;
-      AH_Point**  contour;
+      AH_Point   point;
+      AH_Point   end_point;
+      AH_Point   first_point;
+      AH_Point*  contour;
 
 
       point   = points;
@@ -746,8 +816,8 @@
 
         if ( point <= end_point )
         {
-          AH_Point*  first_touched = point;
-          AH_Point*  cur_touched   = point;
+          AH_Point  first_touched = point;
+          AH_Point  cur_touched   = point;
 
 
           point++;
@@ -792,8 +862,8 @@
         for ( point = points; point < point_limit; point++ )
           point->y = point->u;
 
-        touch_flag = ah_flag_touch_x;
-        ah_setup_uv( outline, ah_uv_ox );
+        touch_flag = AH_FLAG_TOUCH_X;
+        ah_setup_uv( outline, AH_UV_OX );
       }
       else
       {
@@ -808,8 +878,8 @@
 #endif /* !AH_OPTION_NO_WEAK_INTERPOLATION */
 
 
-  FT_LOCAL_DEF void
-  ah_hinter_align_points( AH_Hinter*  hinter )
+  FT_LOCAL_DEF( void )
+  ah_hinter_align_points( AH_Hinter  hinter )
   {
     ah_hinter_align_edge_points( hinter );
 
@@ -836,14 +906,14 @@
 
   /* scale and fit the global metrics */
   static void
-  ah_hinter_scale_globals( AH_Hinter*  hinter,
-                           FT_Fixed    x_scale,
-                           FT_Fixed    y_scale )
+  ah_hinter_scale_globals( AH_Hinter  hinter,
+                           FT_Fixed   x_scale,
+                           FT_Fixed   y_scale )
   {
-    FT_Int            n;
-    AH_Face_Globals*  globals = hinter->globals;
-    AH_Globals*       design = &globals->design;
-    AH_Globals*       scaled = &globals->scaled;
+    FT_Int           n;
+    AH_Face_Globals  globals = hinter->globals;
+    AH_Globals       design = &globals->design;
+    AH_Globals       scaled = &globals->scaled;
 
 
     /* copy content */
@@ -856,8 +926,11 @@
     for ( n = 0; n < design->num_heights; n++ )
       scaled->heights[n] = FT_MulFix( design->heights[n], y_scale );
 
+    scaled->stds[0] = ( design->num_widths  > 0 ) ? scaled->widths[0]  : 32000;
+    scaled->stds[1] = ( design->num_heights > 0 ) ? scaled->heights[0] : 32000;
+
     /* scale the blue zones */
-    for ( n = 0; n < ah_blue_max; n++ )
+    for ( n = 0; n < AH_BLUE_MAX; n++ )
     {
       FT_Pos  delta, delta2;
 
@@ -889,7 +962,7 @@
 
 
   static void
-  ah_hinter_align( AH_Hinter*  hinter )
+  ah_hinter_align( AH_Hinter  hinter )
   {
     ah_hinter_align_edge_points( hinter );
     ah_hinter_align_points( hinter );
@@ -897,8 +970,8 @@
 
 
   /* finalize a hinter object */
-  FT_LOCAL_DEF void
-  ah_hinter_done( AH_Hinter*  hinter )
+  FT_LOCAL_DEF( void )
+  ah_hinter_done( AH_Hinter  hinter )
   {
     if ( hinter )
     {
@@ -914,25 +987,25 @@
       hinter->globals = 0;
       hinter->face    = 0;
 
-      FREE( hinter );
+      FT_FREE( hinter );
     }
   }
 
 
   /* create a new empty hinter object */
-  FT_LOCAL_DEF FT_Error
-  ah_hinter_new( FT_Library   library,
-                 AH_Hinter**  ahinter )
+  FT_LOCAL_DEF( FT_Error )
+  ah_hinter_new( FT_Library  library,
+                 AH_Hinter  *ahinter )
   {
-    AH_Hinter*  hinter = 0;
-    FT_Memory   memory = library->memory;
-    FT_Error    error;
+    AH_Hinter  hinter = 0;
+    FT_Memory  memory = library->memory;
+    FT_Error   error;
 
 
     *ahinter = 0;
 
     /* allocate object */
-    if ( ALLOC( hinter, sizeof ( *hinter ) ) )
+    if ( FT_NEW( hinter ) )
       goto Exit;
 
     hinter->memory = memory;
@@ -956,17 +1029,17 @@
 
 
   /* create a face's autohint globals */
-  FT_LOCAL_DEF FT_Error
-  ah_hinter_new_face_globals( AH_Hinter*   hinter,
-                              FT_Face      face,
-                              AH_Globals*  globals )
+  FT_LOCAL_DEF( FT_Error )
+  ah_hinter_new_face_globals( AH_Hinter   hinter,
+                              FT_Face     face,
+                              AH_Globals  globals )
   {
-    FT_Error          error;
-    FT_Memory         memory = hinter->memory;
-    AH_Face_Globals*  face_globals;
+    FT_Error         error;
+    FT_Memory        memory = hinter->memory;
+    AH_Face_Globals  face_globals;
 
 
-    if ( ALLOC( face_globals, sizeof ( *face_globals ) ) )
+    if ( FT_NEW( face_globals ) )
       goto Exit;
 
     hinter->face    = face;
@@ -988,22 +1061,22 @@
 
 
   /* discard a face's autohint globals */
-  FT_LOCAL_DEF void
-  ah_hinter_done_face_globals( AH_Face_Globals*  globals )
+  FT_LOCAL_DEF( void )
+  ah_hinter_done_face_globals( AH_Face_Globals  globals )
   {
     FT_Face    face   = globals->face;
     FT_Memory  memory = face->memory;
 
 
-    FREE( globals );
+    FT_FREE( globals );
   }
 
 
   static FT_Error
-  ah_hinter_load( AH_Hinter*  hinter,
-                  FT_UInt     glyph_index,
-                  FT_UInt     load_flags,
-                  FT_UInt     depth )
+  ah_hinter_load( AH_Hinter  hinter,
+                  FT_UInt    glyph_index,
+                  FT_Int32   load_flags,
+                  FT_UInt    depth )
   {
     FT_Face           face     = hinter->face;
     FT_GlyphSlot      slot     = face->glyph;
@@ -1011,12 +1084,8 @@
     FT_Fixed          x_scale  = face->size->metrics.x_scale;
     FT_Fixed          y_scale  = face->size->metrics.y_scale;
     FT_Error          error;
-    AH_Outline*       outline  = hinter->glyph;
-    AH_Loader*        gloader  = hinter->loader;
-    FT_Bool           no_horz_hints = FT_BOOL(
-                        ( load_flags & AH_HINT_NO_HORZ_EDGES ) != 0 );
-    FT_Bool           no_vert_hints = FT_BOOL(
-                        ( load_flags & AH_HINT_NO_VERT_EDGES ) != 0 );
+    AH_Outline        outline  = hinter->glyph;
+    AH_Loader         gloader  = hinter->loader;
 
 
     /* load the glyph */
@@ -1031,6 +1100,7 @@
     {
       FT_Matrix  imatrix;
 
+
       imatrix              = internal->glyph_matrix;
       hinter->trans_delta  = internal->glyph_delta;
       hinter->trans_matrix = imatrix;
@@ -1045,7 +1115,7 @@
 
     switch ( slot->format )
     {
-    case ft_glyph_format_outline:
+    case FT_GLYPH_FORMAT_OUTLINE:
 
       /* translate glyph outline if we need to */
       if ( hinter->transformed )
@@ -1068,14 +1138,14 @@
       if ( error )
         goto Exit;
 
-      MEM_Copy( gloader->current.extra_points, slot->outline.points,
-                slot->outline.n_points * sizeof ( FT_Vector ) );
+      FT_MEM_COPY( gloader->current.extra_points, slot->outline.points,
+                   slot->outline.n_points * sizeof ( FT_Vector ) );
 
-      MEM_Copy( gloader->current.outline.contours, slot->outline.contours,
-                slot->outline.n_contours * sizeof ( short ) );
+      FT_MEM_COPY( gloader->current.outline.contours, slot->outline.contours,
+                   slot->outline.n_contours * sizeof ( short ) );
 
-      MEM_Copy( gloader->current.outline.tags, slot->outline.tags,
-                slot->outline.n_points * sizeof ( char ) );
+      FT_MEM_COPY( gloader->current.outline.tags, slot->outline.tags,
+                   slot->outline.n_points * sizeof ( char ) );
 
       gloader->current.outline.n_points   = slot->outline.n_points;
       gloader->current.outline.n_contours = slot->outline.n_contours;
@@ -1099,14 +1169,14 @@
       /* perform feature detection */
       ah_outline_detect_features( outline );
 
-      if ( !no_horz_hints )
+      if ( hinter->do_vert_hints )
       {
         ah_outline_compute_blue_edges( outline, hinter->globals );
         ah_outline_scale_blue_edges( outline, hinter->globals );
       }
 
       /* perform alignment control */
-      ah_hinter_hint_edges( hinter, no_horz_hints, no_vert_hints );
+      ah_hinter_hint_edges( hinter );
       ah_hinter_align( hinter );
 
       /* now save the current outline into the loader's current table */
@@ -1115,10 +1185,10 @@
       /* we now need to hint the metrics according to the change in */
       /* width/positioning that occured during the hinting process  */
       {
-        FT_Pos    old_advance, old_rsb, old_lsb, new_lsb;
-        AH_Edge*  edge1 = outline->vert_edges;     /* leftmost edge  */
-        AH_Edge*  edge2 = edge1 +
-                          outline->num_vedges - 1; /* rightmost edge */
+        FT_Pos   old_advance, old_rsb, old_lsb, new_lsb;
+        AH_Edge  edge1 = outline->vert_edges;     /* leftmost edge  */
+        AH_Edge  edge2 = edge1 +
+                         outline->num_vedges - 1; /* rightmost edge */
 
 
         old_advance = hinter->pp2.x;
@@ -1138,11 +1208,11 @@
       ah_loader_add( gloader );
       break;
 
-    case ft_glyph_format_composite:
+    case FT_GLYPH_FORMAT_COMPOSITE:
       {
-        FT_UInt       nn, num_subglyphs = slot->num_subglyphs;
-        FT_UInt       num_base_subgs, start_point;
-        FT_SubGlyph*  subglyph;
+        FT_UInt      nn, num_subglyphs = slot->num_subglyphs;
+        FT_UInt      num_base_subgs, start_point;
+        FT_SubGlyph  subglyph;
 
 
         start_point   = gloader->base.outline.n_points;
@@ -1152,8 +1222,8 @@
         if ( error )
           goto Exit;
 
-        MEM_Copy( gloader->current.subglyphs, slot->subglyphs,
-                  num_subglyphs * sizeof ( FT_SubGlyph ) );
+        FT_MEM_COPY( gloader->current.subglyphs, slot->subglyphs,
+                     num_subglyphs * sizeof ( FT_SubGlyph ) );
 
         gloader->current.num_subglyphs = num_subglyphs;
         num_base_subgs = gloader->base.num_subglyphs;
@@ -1315,7 +1385,7 @@
         goto Exit;
 
       slot->outline = slot->internal->loader->base.outline;
-      slot->format  = ft_glyph_format_outline;
+      slot->format  = FT_GLYPH_FORMAT_OUTLINE;
     }
 
 #ifdef DEBUG_HINTER
@@ -1328,18 +1398,19 @@
 
 
   /* load and hint a given glyph */
-  FT_LOCAL_DEF FT_Error
-  ah_hinter_load_glyph( AH_Hinter*    hinter,
+  FT_LOCAL_DEF( FT_Error )
+  ah_hinter_load_glyph( AH_Hinter     hinter,
                         FT_GlyphSlot  slot,
                         FT_Size       size,
                         FT_UInt       glyph_index,
-                        FT_Int        load_flags )
+                        FT_Int32      load_flags )
   {
-    FT_Face           face         = slot->face;
-    FT_Error          error;
-    FT_Fixed          x_scale      = size->metrics.x_scale;
-    FT_Fixed          y_scale      = size->metrics.y_scale;
-    AH_Face_Globals*  face_globals = FACE_GLOBALS( face );
+    FT_Face          face         = slot->face;
+    FT_Error         error;
+    FT_Fixed         x_scale      = size->metrics.x_scale;
+    FT_Fixed         y_scale      = size->metrics.y_scale;
+    AH_Face_Globals  face_globals = FACE_GLOBALS( face );
+    FT_Render_Mode   hint_mode    = FT_LOAD_TARGET_MODE(load_flags);
 
 
     /* first of all, we need to check that we're using the correct face and */
@@ -1367,6 +1438,25 @@
 
     ah_loader_rewind( hinter->loader );
 
+    /* reset hinting flags according to load flags and current render target */
+    hinter->do_horz_hints = !FT_BOOL( load_flags & FT_LOAD_NO_AUTOHINT );
+    hinter->do_vert_hints = !FT_BOOL( load_flags & FT_LOAD_NO_AUTOHINT );
+
+#ifdef DEBUG_HINTER
+    hinter->do_horz_hints = !ah_debug_disable_vert;  /* not a bug, the meaning */
+    hinter->do_vert_hints = !ah_debug_disable_horz;  /* of h/v is inverted!    */
+#endif
+
+    /* we snap the width of vertical stems for the monochrome and         */
+    /* horizontal LCD rendering targets only.  Corresponds to X snapping. */
+    hinter->do_horz_snapping = FT_BOOL( hint_mode == FT_RENDER_MODE_MONO ||
+                                        hint_mode == FT_RENDER_MODE_LCD  );
+
+    /* we snap the width of horizontal stems for the monochrome and     */
+    /* vertical LCD rendering targets only.  Corresponds to Y snapping. */
+    hinter->do_vert_snapping = FT_BOOL( hint_mode == FT_RENDER_MODE_MONO   ||
+                                        hint_mode == FT_RENDER_MODE_LCD_V  );
+
 #if 1
     load_flags  = FT_LOAD_NO_SCALE
                 | FT_LOAD_IGNORE_TRANSFORM ;
@@ -1382,19 +1472,19 @@
 
 
   /* retrieve a face's autohint globals for client applications */
-  FT_LOCAL_DEF void
-  ah_hinter_get_global_hints( AH_Hinter*  hinter,
-                              FT_Face     face,
-                              void**      global_hints,
-                              long*       global_len )
+  FT_LOCAL_DEF( void )
+  ah_hinter_get_global_hints( AH_Hinter  hinter,
+                              FT_Face    face,
+                              void**     global_hints,
+                              long*      global_len )
   {
-    AH_Globals*  globals = 0;
-    FT_Memory    memory  = hinter->memory;
-    FT_Error     error;
+    AH_Globals  globals = 0;
+    FT_Memory   memory  = hinter->memory;
+    FT_Error    error;
 
 
     /* allocate new master globals */
-    if ( ALLOC( globals, sizeof ( *globals ) ) )
+    if ( FT_NEW( globals ) )
       goto Fail;
 
     /* compute face globals if needed */
@@ -1412,21 +1502,21 @@
     return;
 
   Fail:
-    FREE( globals );
+    FT_FREE( globals );
 
     *global_hints = 0;
     *global_len   = 0;
   }
 
 
-  FT_LOCAL_DEF void
-  ah_hinter_done_global_hints( AH_Hinter*  hinter,
-                               void*       global_hints )
+  FT_LOCAL_DEF( void )
+  ah_hinter_done_global_hints( AH_Hinter  hinter,
+                               void*      global_hints )
   {
     FT_Memory  memory = hinter->memory;
 
 
-    FREE( global_hints );
+    FT_FREE( global_hints );
   }
 
 
